/*
 *  Video stacking program with change detection.
 *
 *  Panu Lahtinen, pnuu+finsprite@iki.fi
 */

#include "finsprite.h"

void init_peak_hold(void){
  peak_hold->img_buffer = calloc(current_frame.npix,sizeof(*(peak_hold->img_buffer)));
  peak_hold->channels = current_frame.channels;
  peak_hold->width = current_frame.width;
  peak_hold->height = current_frame.height;
  peak_hold->npix = current_frame.npix;
  peak_hold->timestamp = current_frame.timestamp;
  peak_hold->usec = current_frame.usec;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

void init_min_hold(void){
  int i = 0;
  min_hold->img_buffer = malloc(current_frame.npix*sizeof(*(min_hold->img_buffer)));
  for(i=0;i<current_frame.npix;i++){
    min_hold->img_buffer[i] = 255;
  }
  min_hold->channels = current_frame.channels;
  min_hold->width = current_frame.width;
  min_hold->height = current_frame.height;
  min_hold->npix = current_frame.npix;
  min_hold->timestamp = current_frame.timestamp;
  min_hold->usec = current_frame.usec;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

void init_latest_img(void){
  latest_img->img_buffer = malloc(current_frame.npix*sizeof(*(latest_img->img_buffer)));
  latest_img->channels = current_frame.channels;
  latest_img->width = current_frame.width;
  latest_img->height = current_frame.height;
  latest_img->npix = current_frame.npix;
  latest_img->timestamp = current_frame.timestamp;
  latest_img->usec = current_frame.usec;
  latest_img_timestamp = current_frame.timestamp;
  latest_img_usec = current_frame.usec;
}

void init_md_arrays(void){
  int i;
  /* Check motion detection mask dimensions */
  if(md_mask_width != current_frame.width || 
     md_mask_height != current_frame.height){
    if(md_mask != NULL){
      free(md_mask);
      md_mask = NULL;
    }
  }
  
  /* If there is no mask or it was wrong size, generate dummy mask */
  if(md_mask == NULL){
    md_mask = (unsigned char*)malloc(current_frame.width*current_frame.height*sizeof(unsigned char));
    for(i=0;i<current_frame.width*current_frame.height;i++){ md_mask[i] = 1; }
    md_mask_width = current_frame.width;
    md_mask_height = current_frame.height;
  }
    
  /* Allocate all motion detection frames in one go */
  if(md_frames[0].img_buffer == NULL){
    for(i=0; i<img_in_md_buffer; i++){
      md_frames[i].img_buffer = 
	malloc(current_frame.npix*sizeof(*md_frames[i].img_buffer));
      md_frames[i].luminance = NULL; /* No need for this field here */
      md_frames[i].channels = current_frame.channels;
      md_frames[i].width = current_frame.width;
      md_frames[i].height = current_frame.height;
      md_frames[i].npix = current_frame.npix;
    }
  }

  /* Motion detection background arrays */
  md_median1 = (unsigned char*)malloc(current_frame.width*current_frame.height*sizeof(unsigned char));
  md_median2 = (unsigned char*)malloc(current_frame.width*current_frame.height*sizeof(unsigned char));
  
  /* Array of pixel luminances. These values are used in motion
     detection and in calculation of motion detection thresholds for
     each pixel */
  md_luminance1 = (unsigned char**)malloc(img_in_md_buffer*sizeof(unsigned char*));
  md_luminance2 = (unsigned char**)malloc(img_in_md_buffer*sizeof(unsigned char*));
  for(i=0; i<img_in_md_buffer; i++){
    md_luminance1[i] = (unsigned char*)malloc(current_frame.width*
					      current_frame.height*
					      sizeof(unsigned char));
    md_luminance2[i] = (unsigned char*)malloc(current_frame.width*
					      current_frame.height*
					      sizeof(unsigned char));
  }
  
  /* Motion detection histograms */
  md_ts_histogram1 = (int*)malloc(256*sizeof(int));
  md_ts_histogram2 = (int*)malloc(256*sizeof(int));

  /* Point pointers to the arrays currently in use */
  md_median = md_median1;
  md_luminance = md_luminance1;

  md_ts_histogram = md_ts_histogram1;
  
}

void init_yuyv_lookups(void){
  int i, j, k;
  if(yuyv2r == NULL){

    /* Y-direction */
    yuyv2r = (unsigned char**)malloc(256*sizeof(unsigned char*));
    yuyv2g = (unsigned char***)malloc(256*sizeof(unsigned char*));
    yuyv2b = (unsigned char**)malloc(256*sizeof(unsigned char*));

    for(i=0; i<256; i++){
      yuyv2r[i] = (unsigned char*)malloc(256*sizeof(unsigned char));
      yuyv2g[i] = (unsigned char**)malloc(256*sizeof(unsigned char*));
      yuyv2b[i] = (unsigned char*)malloc(256*sizeof(unsigned char));

      for(j=0; j<256; j++){
	yuyv2g[i][j] = (unsigned char*)malloc(256*sizeof(unsigned char));
      }
    }

    /* Calculate lookup tables for each color */
    for(i=0; i<256; i++){
      for(j=0; j<256; j++){
	/* R = limit(Y + 1.4075 * (V - 128)) */
	yuyv2r[i][j] = limit(i + 1.4075 * (j - 128));
	/* B = limit(Y + 1.7790 * (U - 128)) */
	yuyv2b[i][j] = limit(i + 1.7790 * (j - 128));

	for(k=0; k<256; k++){
	  /* G = limit(Y + 0.3455 * (U - 128) - (0.7169 * (V - 128))) */
	  yuyv2g[i][j][k] = limit(i + 0.3455 * (j - 128) - 0.7169 * (k - 128));
	}
      }
    }
  }
}

void *md_process(void *ptr){

  int i = 0, j = 0, n = 0;

  char *tmp = NULL;
  int *hist = NULL;

  md_process_on = 1;

  if(md_detection != -1){
    md_images_since_detection++;
  }

  /* Check if the current daily motion detection counter is up-to-date */
  /* Should this be here??? */
  if(md_trig_number == 0){
    char tmp3[FNAME_SIZE] = {'\0'};
    make_fname(tmp3,current_frame.timestamp,current_frame.usec,DAY_DIR);
    md_trig_number = find_md_folder_num(tmp3);
  }
  
  /* Update the timestamp */
  md_frames[md_i].timestamp = current_frame.timestamp;
  md_frames[md_i].usec = current_frame.usec;
  
  /* Copy current image buffer to ring memory */
  j = 0;
  for(i=0; i<current_frame.npix; i++){
    md_frames[md_i].img_buffer[i] = current_frame.img_buffer[i];
    /* Copy also luminance data */
    if(current_frame.channels == 3){
      if(i%3 == 0){
	/* Luminance from the original frame */
	md_luminance[md_i][j] = current_frame.luminance[j];
	j++;
      }
    } else {
      md_luminance[md_i][i] = current_frame.img_buffer[i];
    }
			    
  }

  /* Detect rapid changes in illumination (motion detection) */
  /* XXXXX change to use ts-tl histogram */
  if(md_buffer_ready && md_detection == -1 && md_references_ok){

    /* XXXX calculate ts-tl histogram for current frame */

    hist = (int*)calloc(256,sizeof(int));
    calc_ts_histogram(hist, md_median, md_luminance[md_i], 
		      current_frame.width*current_frame.height);

    /* XXXX run comparison function: if return value > -1, ->
       detection, write log, etc */
    if((n = histogram_compare(hist, md_ts_histogram, md_min_tl)) > 0){
      if(md_detection == -1){
	md_detection = md_i;
	md_images_since_detection = 0;
      }
      tmp = make_time_string(current_frame.timestamp,current_frame.usec,TRIG);
      /* Get the date for the frame that triggers the saving. All the 
	 frames from this event are saved to same folder, no matter what 
	 date. */
      sscanf(tmp,"%d-%d-%d",&md_trig_year,&md_trig_month,&md_trig_day);

      if(save_log){
	if(md_log_fname == NULL){
	  md_log_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
	  memset(md_log_fname,'\0',FNAME_SIZE);
	}
	
	char tmp2[FNAME_SIZE] = {'\0'};
	
	make_fname(tmp2,current_frame.timestamp,current_frame.usec,MD_LOG_NAME);
	
	/* New file to be opened */
	if(md_log_fname[0] == '\0' || strcmp(tmp2, md_log_fname) != 0){
	  if(md_log_fid != NULL){
	    fclose(md_log_fid); md_log_fid = NULL;
	  }
	  
	  strcpy(md_log_fname,tmp2);
	  /* Add header if this is the first time we open the file
	   */
	  if(file_exists(md_log_fname)){
	    md_log_fid = fopen(md_log_fname,"a");
	  } else { // New file, add header and reset MD event counter
	    md_log_fid = fopen(md_log_fname,"a");
	    fprintf(md_log_fid,"# year, month, day, hour, minute, second, detection level, detection size, required minimum detection level, required detection size, images saved, images saved before event, images saved after event, event number\n");
	    md_trig_number = 0;
	  }
	}
	
	if(tmp != NULL){ free(tmp); }
	
	tmp = make_time_string(current_frame.timestamp,current_frame.usec,MD_LOG_DATA);
	
	/* Now the file should be open until program exits or the 
	   day changes */
	
	int n_before = md_percentage * img_in_md_buffer / 100;
	
	fprintf(md_log_fid,"%s,%d,%d,%d,%d,%d,%d,%d,%d\n",
		tmp,
		n, /* Detection level */
		hist[n], /* Detection size */
		md_min_tl, /* Minimum allowed detection level */
		md_ts_histogram[n], /* Detection size that was required */
		img_in_md_buffer,
		n_before,
		img_in_md_buffer-n_before-1,
		md_trig_number);
	fflush(md_log_fid);
	free(tmp);
      }
    } 
  }
  
  // Increment index where to store image
  md_i++;
  // If we are using the last slot of the image buffer, start from zero
  if(md_i >= img_in_md_buffer){
    md_i = 0;
    md_buffer_ready = 1;
  }
  
  md_process_on = 0;
  
  pthread_exit(NULL);

}

void *calc_md_references(void *ptr){

  printf("calc_md_references\n");

  md_reference_calc_on = 1;

  int n_pix = current_frame.width*current_frame.height;
  int n_imgs = img_in_md_buffer;

  int i, j;
  unsigned char arr[n_imgs];

  int **md_ts_histograms = NULL;

  /* Calculate median value for each pixel */
  for(i=0; i<n_pix; i++){
    if(md_mask[i] == 1){
      for(j=0; j<n_imgs; j++){
	arr[j] = md_luminance_calc[j][i];
      }
      md_median_calc[i] = median_uint8(arr, n_imgs);
    }
  }

  /* Calculate trigger size histograms for each image */
  md_ts_histograms = (int**)malloc(img_in_md_buffer*sizeof(int*));
  for(i=0; i<img_in_md_buffer; i++){
    md_ts_histograms[i] = (int*)calloc(256,sizeof(int));
    calc_ts_histogram(md_ts_histograms[i], md_median_calc, 
		      md_luminance_calc[i], 
		      current_frame.width*current_frame.height);
  }

  /* Calculate median of the histograms */
  int *hist_ptr = NULL;
  hist_ptr = (int*)malloc(img_in_md_buffer*sizeof(int));
  for(i=0; i<256; i++){
    for(j=0; j<img_in_md_buffer; j++){
      hist_ptr[j] = md_ts_histograms[j][i];
    }
    md_ts_histogram_calc[i] = median_int(hist_ptr, img_in_md_buffer);
  }

  /* Modify the histogram so that it is shifted up and right by N
     steps/values */
  for(i=255; i>md_histogram_adj; i--){
    md_ts_histogram_calc[i] = md_ts_histogram_calc[i-md_histogram_adj] + md_histogram_adj;
  }

  /* Wait until it is safe to swap the pointers */
  while(md_process_on && !md_references_ok){printf("hop\n");}
  /* Now the calculations are done, swap these in the place of previous values */
  if(md_median == md_median1){
    md_median = md_median2;
    md_median_calc = md_median1;
    md_ts_histogram = md_ts_histogram2;
    md_ts_histogram_calc = md_ts_histogram1;
  } else {
    md_median = md_median1;
    md_median_calc = md_median2;
    md_ts_histogram = md_ts_histogram1;
    md_ts_histogram_calc = md_ts_histogram2;
  }

  time(&md_background_update_time);

  /* Release local variables */
  free(hist_ptr);
  for(i=0; i<img_in_md_buffer; i++){
    free(md_ts_histograms[i]);
  }
  free(md_ts_histograms);

  md_reference_calc_on = 0;
  md_references_ok = 1;
  printf("calc_md_references done\n");

  pthread_exit(NULL);

}

void *md_save_images(void* ptr){

  char *fname = NULL;
  fname = (char*)malloc(FNAME_SIZE*sizeof(char));

  int i = 0, j = 0;

  unsigned char *md_peak_hold = (unsigned char*)calloc(md_frames[0].npix,sizeof(unsigned char));

  /* Disable motion detection until the ring memory is being saved */
  md_saving_images = 1;

  /* Save reference image */
  memset(fname,'\0',FNAME_SIZE);
  make_fname(fname,-1,-1,REFERENCE);
  writeImage(fname, 1, current_frame.width, current_frame.height, md_median, "spam!");


  /* Save ring memory to images on disc */
  if(!md_test){
    printf("Saving ring memory..\n");
  }
  for(i=0; i<img_in_md_buffer; i++){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname,md_frames[i].timestamp, md_frames[i].usec, TRIG);
    if(!md_test){
      if(verbose){
	printf("[%d] %s saved from ring memory\n",i,fname);
      }
      writeImage(fname, 
		 md_frames[i].channels,
		 md_frames[i].width,
		 md_frames[i].height,
		 md_frames[i].img_buffer,
		 "Finsprite motion detection at work");
      for(j=0; j<md_frames[i].npix; j++){
	if(md_peak_hold[j] < md_frames[i].img_buffer[j]){
	  md_peak_hold[j] = md_frames[i].img_buffer[j];
	}
      }
    } else {
      if(verbose){
	printf("[%d] %s read from ring memory\n",i,fname);
      }
    }
  }

  if(!md_test){
    memset(fname,'\0',FNAME_SIZE);
    //i = md_detection - (img_in_md_buffer-1)/2;
    //if(i < 0){ i += img_in_md_buffer; }

    make_fname(fname,md_frames[md_i].timestamp, md_frames[md_i].usec, TRIG_PEAK);
    writeImage(fname, 
	       md_frames[0].channels,
	       md_frames[0].width,
	       md_frames[0].height,
	       md_peak_hold,
	       "Finsprite motion detection at work");

    printf("%d images from ring memory saved.\n",img_in_md_buffer);
  }

  md_trig_number++;

  md_images_since_detection = -1;
  md_detection = -1;

  /* Start to fill the motion detection ring memory from the beginning */
  md_i = 0;
  md_buffer_ready = 0;

  /* Enable motion detection */
  md_saving_images = 0;

  free(fname); fname = NULL;

  pthread_exit(NULL);

}

void *process_stack(void* ptr){

  int i = 0;
  unsigned long min = ULONG_MAX;

  /* Initialize stacks, if they are empty*/
  if(use_max || use_dif){
    if(peak_hold->img_buffer == NULL){
      init_peak_hold();
    }
  }

  if(use_latest){
    if(latest_img->img_buffer == NULL){
      init_latest_img();
    }
  }
  
  if(use_min || use_dif){
    if(min_hold->img_buffer == NULL){
      init_min_hold();
    }
  }
  
  if(use_ave){
    if(ave_stack->img_buffer == NULL){
      ave_stack->img_buffer = calloc(current_frame.npix, sizeof(*(ave_stack->img_buffer)));
      ave_stack->channels = current_frame.channels;
      ave_stack->width = current_frame.width;
      ave_stack->height = current_frame.height;
      ave_stack->npix = current_frame.npix;
      ave_stack->timestamp = current_frame.timestamp;
      ave_stack->usec = current_frame.usec;
      ave_stack->max_val = 0;
      ave_stack->min_val = ULONG_MAX;
      stack_timestamp = current_frame.timestamp;
      stack_usec = current_frame.usec;
    }
  }

  /* Update stacks */
  for(i=0;i<current_frame.npix;i++){

    if(use_max || use_dif){
      if(current_frame.img_buffer[i] > peak_hold->img_buffer[i]){
	peak_hold->img_buffer[i] = current_frame.img_buffer[i];
      }
    }

    if(use_latest){
      if(current_frame.img_buffer[i] > latest_img->img_buffer[i]){
	latest_img->img_buffer[i] = current_frame.img_buffer[i];
      }
    }

    if(use_min || use_dif){
      if(current_frame.img_buffer[i] < min_hold->img_buffer[i]){
	min_hold->img_buffer[i] = current_frame.img_buffer[i];
      }
    }

    if(use_ave){
      ave_stack->img_buffer[i] += current_frame.img_buffer[i];

      if(ave_stack->img_buffer[i] > ave_stack->max_val){
	ave_stack->max_val = ave_stack->img_buffer[i];
      }
      
      if(ave_stack->img_buffer[i] < min){
	min = ave_stack->img_buffer[i];
      }
    }
  }

  if(use_ave){
    ave_stack->num_imgs++;

    if(min > ave_stack->min_val){
      ave_stack->min_val = min;
    }
  }

  pthread_exit(NULL);

}

static int process_frame(const unsigned char *p, time_t timestamp, long usec){

  unsigned char Y1=0, U=0, Y2=0, V=0;

  //  int x=0, y=0;
  
  int i = 0, j = 0, k = 0;

  int maximum = 0;

  current_frame.channels = number_of_channels;
  current_frame.width = fmt.fmt.pix.width;
  current_frame.height = fmt.fmt.pix.height;
  current_frame.npix = current_frame.channels * current_frame.width * current_frame.height;
  current_frame.timestamp = timestamp;
  current_frame.usec = usec;

  if(current_frame.img_buffer == NULL){
    current_frame.img_buffer = (unsigned char*)malloc(current_frame.npix*sizeof(unsigned char*));
    if(current_frame.channels == 3 || use_trig){
      current_frame.luminance = (unsigned char*)malloc(current_frame.width*
						       current_frame.height*
						       sizeof(unsigned char*));
    }
  }

  //while(j+3 < (fmt.fmt.pix.width * fmt.fmt.pix.height)){
  /* Assume YUYV format. RGB and GREY should be added? */
  for(i = 0; i < fmt.fmt.pix.sizeimage; i+=4){
    
    switch(fmt.fmt.pix.pixelformat){
    case V4L2_PIX_FMT_GREY:
      Y1 = p[j];
      U = 0;
      Y2 = p[j+2];
      V = 0;
      break;
    case V4L2_PIX_FMT_UYVY:
      U = p[i];
      Y1 = p[i+1];
      V = p[i+2];
      Y2 = p[i+3];
      break;
    case V4L2_PIX_FMT_YUYV:
      Y1 = p[i];
      U = p[i+1];
      Y2 = p[i+2];
      V = p[i+3];
    default:
      break;
    }
    if(number_of_channels == 3){
      /* R1 */
      //current_frame.img_buffer[j+0] = limit(Y1 + 1.4075 * (V - 128));
      current_frame.img_buffer[j+0] = yuyv2r[Y1][V];
      /* G1 */
      //current_frame.img_buffer[j+1] = limit(Y1 + 0.3455 * (U - 128) - (0.7169 * (V - 128)));
      current_frame.img_buffer[j+1] = yuyv2g[Y1][U][V];
      /* B1 */
      //current_frame.img_buffer[j+2] = limit(Y1 + 1.7790 * (U - 128));
      current_frame.img_buffer[j+2] = yuyv2b[Y1][U];
      
      /* R2 */
      //current_frame.img_buffer[j+3] = limit(Y2 + 1.4075 * (V - 128));
      current_frame.img_buffer[j+3] = yuyv2r[Y2][V];
      /* G2 */
      //current_frame.img_buffer[j+4] = limit(Y2 + 0.3455 * (U - 128) - (0.7169 * (V - 128)));
      current_frame.img_buffer[j+4] = yuyv2g[Y2][U][V];
      /* B2 */
      //current_frame.img_buffer[j+5] = limit(Y2 + 1.7790 * (U - 128));
      current_frame.img_buffer[j+5] = yuyv2b[Y2][U];
      
      j += 6;

      /* Store luminance if motion detection is used*/
      if(use_trig){
	current_frame.luminance[k+0] = Y1;
	current_frame.luminance[k+1] = Y2;
	k += 2;
      }

    } else { /* B&W image */
      current_frame.img_buffer[j+0] = Y1;
      current_frame.img_buffer[j+1] = Y2;
      
      j += 2;
    }

    if(Y1+Y2 > maximum){
      maximum = Y1+Y2;
    }
    
  }

  return(maximum);

}

void *save_stacks(void *ptr){

  saving_stacks = 1;

  /* Save stack information to log */
  if(save_log){
    if(stack_log_fname == NULL){
      stack_log_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      memset(stack_log_fname,'\0',FNAME_SIZE);
    }

    char tmp[FNAME_SIZE] = {'\0'};
    char *tmp2 = NULL;

    make_fname(tmp,stack_timestamp,stack_usec,STACK_LOG_NAME);

    /* New file to be opened */
    if(stack_log_fname[0] == '\0' || strcmp(tmp, stack_log_fname) != 0){
      if(stack_log_fid != NULL){
	fclose(stack_log_fid); stack_log_fid = NULL;
      }

      strcpy(stack_log_fname,tmp);
      if(file_exists(stack_log_fname)){
	stack_log_fid = fopen(stack_log_fname,"a");
      } else {
	stack_log_fid = fopen(stack_log_fname,"a");
	fprintf(stack_log_fid,"# start year, start month, start day, start hour, start minute, start second, end year, end month, end day, end hour, end minute, end second, stack length, images in stack, min stack saved, max stack saved, ave stack saved, dif stack saved\n");
      }
    }

    tmp2 = make_time_string(stack_timestamp,stack_usec,STACK_LOG_DATA);
    fprintf(stack_log_fid,"%s,%d,%d,%d,%d,%d,%d\n",tmp2,stack_length,num_img_in_save_stack,use_min,use_max,use_ave,use_dif);
    fflush(stack_log_fid);

    free(tmp2);

  }

  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);  

  if(use_ave){
    save_ave_stack();
  }
  
  if(use_dif){
    save_dif_stack();
  }

  if(use_max){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname, peak_hold_save->timestamp, peak_hold_save->usec, PEAK_HOLD);
    printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
    //stretch(peak_hold->img_buffer, peak_hold->width*peak_hold->height, 0.01, 0.99);
    writeImage(fname, 
	       peak_hold_save->channels,
	       peak_hold_save->width,
	       peak_hold_save->height,
	       peak_hold_save->img_buffer,
	       "Finsprite peak-hold stack");
  }
  
  if(use_max || use_dif){
    free(peak_hold_save->img_buffer); peak_hold_save->img_buffer = NULL;
  }
  
  if(use_min){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname, min_hold_save->timestamp, min_hold_save->usec, MIN_HOLD);
    printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
    //stretch(min_hold->img_buffer, min_hold->width*min_hold->height, 0.01, 0.99);
    writeImage(fname, 
	       min_hold_save->channels,
	       min_hold_save->width,
	       min_hold_save->height,
	       min_hold_save->img_buffer,
	       "Finsprite minimum stack");

    free(min_hold_save->img_buffer); min_hold_save->img_buffer = NULL;
  }
  
  if(use_min || use_dif){
    free(min_hold_save->img_buffer); min_hold_save->img_buffer = NULL;
  }
  
  num_img_in_save_stack = 0;
  //  stack_timestamp = current_frame.timestamp;

  free(fname);

  saving_stacks = 0;

  pthread_exit(NULL);

}

void *save_latest(void *ptr){

  saving_latest = 1;

  printf("Saving %s\n",latest_fname);
  writeImage(latest_fname,
	     latest_img_save->channels,
	     latest_img_save->width,
	     latest_img_save->height,
	     latest_img_save->img_buffer,
	     "Finsprite latest stack");
  
  free(latest_img_save->img_buffer); latest_img_save->img_buffer = NULL;
  
  saving_latest = 0;

  pthread_exit(NULL);

}

static void errno_exit(const char *s){
  fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
  exit(EXIT_FAILURE);
}

static int xioctl(int fd, int request, void *arg){
  int r;
  
  do {
    r = ioctl(fd, request, arg);
  } while(r == -1 && EINTR == errno);
  
  return r;
}

char *make_time_string(time_t timestamp, long usec, int type){

  char *time_string = (char*)malloc(100*sizeof(char));
  char *tmp = NULL;

  if(type < MD_LOG_NAME){
    strftime(time_string,30,"%Y-%m-%d_%H%M%S",gmtime(&(timestamp)));
  }

  switch(type){
  case TRIG:
  case TRIG_PEAK:
    tmp = (char*)malloc(50*sizeof(char));
    sprintf(tmp,".%.6ld",usec);
    strcat(time_string,tmp);
    free(tmp);
    break;
  case MD_LOG_NAME:
  case STACK_LOG_NAME:
    strftime(time_string,30,"%Y-%m-%d",gmtime(&(timestamp)));
    break;
  case MD_LOG_DATA:
  case STACK_LOG_DATA:
    tmp = (char*)malloc(30*sizeof(char));
    strftime(time_string,30,"%Y,%m,%d,%H,%M,%S",gmtime(&(timestamp)));
    sprintf(tmp,".%.6ld",usec);
    strcat(time_string,tmp);
    if(type == STACK_LOG_DATA){
      memset(tmp,'\0',30);
      strftime(tmp,30,",%Y,%m,%d,%H,%M,%S",gmtime(&(stack_end_timestamp)));
      strcat(time_string,tmp);
      memset(tmp,'\0',30);
      sprintf(tmp,".%.6ld",stack_end_usec);
      strcat(time_string,tmp);
    }
    free(tmp);
    break;
  }
  return time_string;
}

void make_fname(char *fname, time_t timestamp, long usec, int type){
  char tmp[50] = {'\0'};
  mode_t dir_mode = S_IRWXU |S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

  if(out_dir != NULL){ strcat(fname,out_dir); strcat(fname,"/"); }

  if(type == TRIG || type == TRIG_PEAK || type == REFERENCE){
    sprintf(tmp,"%d",md_trig_year);
  } else {
    strftime(tmp,50,"%Y",gmtime(&(timestamp)));
  }
  strcat(fname,tmp);
  memset(tmp,'\0',50);

  if(mkdir(fname, dir_mode) < 0 && errno != EEXIST){
    printf("Unable to create %s\nSaving to current direcory: %d\n",fname,errno);
    // Empty path
    memset(fname,'\0',FNAME_SIZE);
  } else {

    if(type == TRIG || type == TRIG_PEAK || type == REFERENCE){
      sprintf(tmp,"/%.2d",md_trig_month);
    } else {
      strftime(tmp,50,"/%m",gmtime(&(timestamp)));
    }
    strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    // Save only images to daily subfolders
    if(type < MD_LOG_NAME){
      if(type == TRIG || type == TRIG_PEAK || type == REFERENCE){
	sprintf(tmp,"/%.2d",md_trig_day);
      } else {
	strftime(tmp,50,"/%d",gmtime(&(timestamp)));
      }
      strcat(fname,tmp); CLEAR(tmp);
      mkdir(fname, dir_mode);
    }

    /* Save motion detected images to event-folders, stacks to respective
       folders
    */
    switch(type){
    case MIN_HOLD:
      strcat(fname,"/min");
      break;
    case PEAK_HOLD:
      strcat(fname,"/max");
      break;
    case AVE:
      strcat(fname,"/ave");
      break;
    case DIF:
      strcat(fname,"/dif");
      break;
    case TRIG:
    case TRIG_PEAK:
    case REFERENCE:
      sprintf(tmp,"/MD%.4d",md_trig_number);
      strcat(fname,tmp);
      break;
    }

    mkdir(fname, dir_mode);
    memset(tmp,'\0',50);

    strcat(fname,"/");

  }

  /* If we want to have the daily image directory, skip filename creation */
  if(type != DAY_DIR){

    if(type != REFERENCE){
      if(img_prefix != NULL){ strcat(fname,img_prefix); strcat(fname,"_"); }
      char *tmp2 = NULL;
      tmp2 = make_time_string(timestamp,usec,type);
      strcat(fname,tmp2); 
      free(tmp2);
    } else {
      strcat(fname,"reference.png");
    }
    
    switch(type){
    case AVE:
      sprintf(tmp,"_UTC_%d_s_ave.png", stack_length);
      break;
    case PEAK_HOLD:
      sprintf(tmp,"_UTC_%d_s_max.png", stack_length);
      break;
    case MIN_HOLD:
      sprintf(tmp,"_UTC_%d_s_min.png", stack_length);
      break;
    case DIF:
      sprintf(tmp,"_UTC_%d_s_dif.png", stack_length);
      break;
    case TRIG:
      sprintf(tmp,"_UTC_md.png");
      break;
    case TRIG_PEAK:
      sprintf(tmp,"_UTC_md_max.png");
      break;
    case MD_LOG_NAME:
      sprintf(tmp,"_md.log");
      break;
    case STACK_LOG_NAME:
      sprintf(tmp,"_stacks.log");
      break;
    case REFERENCE:
      break;
    default:
      sprintf(tmp,".%ld_UTC.png",usec);
      break;
    }
    strcat(fname,tmp);
  }

}

void save_ave_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);

  struct frame out_frame;

  out_frame.img_buffer = calloc(ave_stack_save->npix,sizeof(unsigned char));
  out_frame.channels = ave_stack_save->channels;
  out_frame.width = ave_stack_save->width;
  out_frame.height = ave_stack_save->height;
  out_frame.timestamp = ave_stack_save->timestamp;
  out_frame.usec = ave_stack_save->usec;

  if(ave_stack_save->img_buffer == NULL)
    printf("\n\nave_stack->img_buffer == NULL\n\n");

  for(i=0;i<ave_stack_save->npix;i++){
    out_frame.img_buffer[i] = ave_stack_save->img_buffer[i]/ave_stack_save->num_imgs;
    //out_frame[i] = 255*ave_stack_save->img_buffer[i]/ave_stack_save->max_val;
    //out_frame[i] = 255*(ave_stack_save->img_buffer[i]-ave_stack_save->min_val)/(ave_stack_save->max_val-ave_stack_save->min_val);
  }

  make_fname(fname, ave_stack_save->timestamp, ave_stack_save->usec, AVE);
  //stretch(out_frame, ave_stack_save->width*ave_stack_save->height, 0.02, 0.98);
  printf("Saving %s, %d frames used.\n",fname,ave_stack_save->num_imgs);
  writeImage(fname, 
	     out_frame.channels,
	     out_frame.width,
	     out_frame.height,
	     out_frame.img_buffer,
	     "Finsprite average stack");
  
  free(out_frame.img_buffer);
  free(fname);

  free(ave_stack_save->img_buffer);
  ave_stack_save->img_buffer = NULL;
  ave_stack_save->min_val = ULONG_MAX;
  ave_stack_save->max_val = 0;
  ave_stack_save->num_imgs = 0;
}

void save_dif_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);

  unsigned char *dif = calloc(min_hold_save->npix,sizeof(float));

  struct frame out_frame;
  out_frame.img_buffer = calloc(min_hold_save->npix,sizeof(unsigned char));
  out_frame.channels = min_hold_save->channels;
  out_frame.width = min_hold_save->width;
  out_frame.height = min_hold_save->height;
  out_frame.timestamp = min_hold_save->timestamp;
  out_frame.usec = min_hold_save->usec;

  unsigned char dmin = UCHAR_MAX, dmax = 0;
  for(i=0; i<min_hold_save->npix; i++){
    dif[i] = peak_hold_save->img_buffer[i] - min_hold_save->img_buffer[i];
    if(dif[i] < dmin){ dmin = dif[i]; }
    if(dif[i] > dmax){ dmax = dif[i]; }
  }

  for(i=0;i<min_hold_save->npix;i++){
    out_frame.img_buffer[i] = 255*(dif[i]-dmin)/dmax;
  }
  
  free(dif); dif = NULL;

  make_fname(fname, min_hold_save->timestamp, min_hold_save->usec, DIF);
  printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
  //stretch(out_frame, min_hold->width*min_hold->height, 0.02, 0.98);
  writeImage(fname, 
	     out_frame.channels,
	     out_frame.width,
	     out_frame.height,
	     out_frame.img_buffer,
	     "Finsprite difference stack");
  
  free(out_frame.img_buffer); out_frame.img_buffer = NULL;
  free(fname); fname = NULL;

}

static int read_frame(void){
  struct v4l2_buffer buf;

  time_t frametime;
  
  int status = 0, i, j;
  void *status2 = NULL;

  CLEAR(buf);

  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
    switch(errno) {
    case EAGAIN:
      printf("EAGAIN\n");
      return 0;
      
    case EIO:
      /* Could ignore EIO, see spec. */
      printf("EIO\n");
      /* fall through */
      
    default:
      printf("default\n");
      errno_exit("VIDIOC_DQBUF");
    }
  }
    
  assert(buf.index < n_mmap_buffers);
  
  frametime = buf.timestamp.tv_sec;

  if(frametime > 0){
    if(process_frame(mmap_buffers[buf.index].start, frametime, buf.timestamp.tv_usec) > 0){

      if(use_trig && md_luminance == NULL){
	init_md_arrays();
      } else {

	num_img_in_stack++;

	//      pthread_t stack_thread, trig_thread;
	pthread_attr_t thread_attr;

	int stack_thread_used = 0, trig_thread_used = 0;
	
	pthread_attr_init(&thread_attr);
	pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_JOINABLE);
	
	/* Check statuses of motion detection stuff, if it is enabled */
	if(use_trig){
	  /*XXXXXXXXX don't let this stop updating the ring buffers! */
	  /* If ring memory is ready to be saved, create a new thread for it */
	  if(!md_saving_images && 
	     100.0*md_images_since_detection/img_in_md_buffer >= md_percentage){
	    status = pthread_create(&trig_save_thread, NULL, md_save_images, NULL);
	    if(status){
	      printf("Thread \"md_save_images()\" failed with code %d\n",status);
	    }
	  }

	  /* If motion detection ring buffer is full, and if there is
	     no calculated dynamic threshold data or it is too old,
	     calculate needed matrices */
	  if(md_buffer_ready && !md_reference_calc_on &&
	     md_threshold_update_interval <= difftime(current_frame.timestamp, md_background_update_time)){
	    if(md_median == md_median1){
	      md_median = md_median2;
	      md_median_calc = md_median1;
	      md_luminance = md_luminance2;
	      md_luminance_calc = md_luminance1;
	      md_ts_histogram = md_ts_histogram2;
	      md_ts_histogram_calc = md_ts_histogram1;
	    } else {
	      md_median = md_median1;
	      md_median_calc = md_median2;
	      md_luminance = md_luminance1;
	      md_luminance_calc = md_luminance2;
	      md_ts_histogram = md_ts_histogram1;
	      md_ts_histogram_calc = md_ts_histogram2;
	    }
	    /* Copy existing data to the matrices to be used in motion
	       detection while the threshold calculations are done */
	    if(md_background_update_time > 0){
	      for(i=0; i<current_frame.width*current_frame.height; i++){
		md_median[i] = md_median_calc[i];
		for(j=0; j<img_in_md_buffer; j++){
		  md_luminance[j][i] = md_luminance_calc[j][i];
		}
	      }
	      for(i=0; i<256; i++){
		md_ts_histogram[i] = md_ts_histogram_calc[i];
	      }
	    }

	    status = pthread_create(&md_references_thread, NULL, calc_md_references, NULL);
	  }

	}
	
	if(use_trig && !md_saving_images){
	  trig_thread_used = 1;
	  status = pthread_create(&trig_thread, &thread_attr, md_process, NULL);
	  //status = pthread_create(&md_trig_thread, NULL, md_process, NULL);
	  if(status){
	    printf("Thread \"md_process()\" failed with code %d\n",status);
	  }
	}

	if(use_max || use_min || use_ave || use_dif || use_latest){
	  
	  stack_thread_used = 1;
	  status = pthread_create(&stack_thread, &thread_attr, process_stack, NULL);
	  //status = pthread_create(&stack_thread, NULL, process_stack, NULL);
	  if(status){
	    printf("Thread \"process_stack()\" failed with code %d\n",status);
	  }
	}

	/* Join threads */
      
	pthread_attr_destroy(&thread_attr);
	if(stack_thread_used){
	  status = pthread_join(stack_thread, &status2);
	  if(status) {
	    printf("ERROR; return code from pthread_join() for stack_thread is %d\n", status);
	  }
	}

	if(trig_thread_used){
	  status = pthread_join(trig_thread, &status2);
	  if(status) {
	    printf("ERROR; return code from pthread_join() for stack_thread is %d\n", status);
	  }
	}
      
	if((use_ave || use_min || use_max || use_dif) && stack_length < difftime(current_frame.timestamp,stack_timestamp)){

	  stack_end_timestamp = current_frame.timestamp;
	  stack_end_usec = current_frame.usec;

	  num_img_in_save_stack = num_img_in_stack;
	  num_img_in_stack = 0;

	  /* Swap stack pointers and create thread for stack saving */
	  if(min_hold == &min_hold1){
	    min_hold = &min_hold2;
	    min_hold_save = &min_hold1;
	  } else {
	    min_hold = &min_hold1;
	    min_hold_save = &min_hold2;
	  }
	  if(peak_hold == &peak_hold1){
	    peak_hold = &peak_hold2;
	    peak_hold_save = &peak_hold1;
	  } else {
	    peak_hold = &peak_hold1;
	    peak_hold_save = &peak_hold2;
	  }
	  if(ave_stack == &ave_stack1){
	    ave_stack = &ave_stack2;
	    ave_stack_save = &ave_stack1;
	  } else {
	    ave_stack = &ave_stack1;
	    ave_stack_save = &ave_stack2;
	  }

	  status = pthread_create(&stack_save_thread, NULL, save_stacks, NULL);
	  if(status){
	    printf("Thread \"save_stacks()\" failed with code %d\n",status);
	  }

	}

	if(use_latest && latest_length < difftime(current_frame.timestamp,latest_img_timestamp)){

	  if(latest_img == &latest_img1){
	    latest_img = &latest_img2;
	    latest_img_save = &latest_img1;
	  } else {
	    latest_img = &latest_img1;
	    latest_img_save = &latest_img2;
	  }

	  status = pthread_create(&latest_save_thread, NULL, save_latest, NULL);
	  if(status){
	    printf("Thread \"save_latest()\" failed with code %d\n",status);
	  }

	} // latest.png saving

      }
    } // if process_frame got good data

  } /* if valid frame time was received */

  if(capture_length != 0){
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }

  return 1;
}

static void mainloop(void){
  time_t timenow;

  time(&start_time);
  time(&timenow);

  while(difftime(timenow,start_time) < capture_length) {

    for(;;) {
      fd_set fds;
      struct timeval tv;
      int r;
      
      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      
      /* Timeout. */
      tv.tv_sec = 2;
      tv.tv_usec = 0;
      
      r = select(fd + 1, &fds, NULL, NULL, &tv);
      
      if(r == -1) {
	if(EINTR == errno)
	  continue;
	
	errno_exit("select");
      }
      
      if(r == 0) {
	fprintf(stderr, "select timeout\n");
	exit(EXIT_FAILURE);
      }
      
      if(read_frame())
	break;
      
      /* EAGAIN - continue select loop. */
    }
    time(&timenow);
  }

}

static void stop_capturing(void){
  enum v4l2_buf_type type;
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMOFF, &type) == -1)
    errno_exit("VIDIOC_STREAMOFF");

}

static void start_capturing(void){
  unsigned int i;
  enum v4l2_buf_type type;
  
  for(i = 0; i < n_mmap_buffers; ++i) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = i;
    
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMON, &type) == -1){
    errno_exit("VIDIOC_STREAMON");
  }
}

static void uninit_device(void){
  unsigned int i;
  
  for(i = 0; i < n_mmap_buffers; ++i)
    if(munmap(mmap_buffers[i].start, mmap_buffers[i].length) == -1)
      errno_exit("munmap");
  
  free(mmap_buffers);
}

static void init_mmap(int n){
  struct v4l2_requestbuffers req;
  
  CLEAR(req);
  
  req.count = 2;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s does not support memory mapping\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_REQBUFS");
    }
  }
  
  if(req.count < 2) {
    fprintf(stderr, "Insufficient buffer memory on %s\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  mmap_buffers = calloc(req.count, sizeof(*mmap_buffers));
  
  if(!mmap_buffers) {
    fprintf(stderr, "Out of memory\n");
    exit(EXIT_FAILURE);
  }
  
  for(n_mmap_buffers = 0; n_mmap_buffers < req.count; ++n_mmap_buffers) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n_mmap_buffers;
    
    if(xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
      errno_exit("VIDIOC_QUERYBUF");
    
    mmap_buffers[n_mmap_buffers].length = buf.length;
    mmap_buffers[n_mmap_buffers].start =
      mmap(NULL /* start anywhere */,
	   buf.length,
	   PROT_READ | PROT_WRITE /* required */,
	   MAP_SHARED /* recommended */,
	   fd, buf.m.offset);
    
    if(MAP_FAILED == mmap_buffers[n_mmap_buffers].start)
      errno_exit("mmap");
  }
}

static void init_device(int img_in_mmap_buffer){
  struct v4l2_capability cap;
  struct v4l2_cropcap cropcap;
  struct v4l2_crop crop;
  unsigned int min;
  
  if(xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s is no V4L2 device\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_QUERYCAP");
    }
  }
  
  if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    fprintf(stderr, "%s is no video capture device\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
    fprintf(stderr, "%s does not support streaming i/o\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  /* Select video input, video standard and tune here. */
  
  CLEAR(cropcap);

  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0) {
    crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c = cropcap.defrect; /* reset to default */
    
    if(xioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
      switch(errno) {
      case EINVAL:
	/* Cropping not supported. */
	break;
      default:
	/* Errors ignored. */
	break;
      }
    }
  } else {
    /* Errors ignored. */
  }
    
  CLEAR(fmt);
  
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = 720;
  fmt.fmt.pix.height = 576;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

  if(xioctl(fd, VIDIOC_S_FMT, &fmt) == -1)
    errno_exit("VIDIOC_S_FMT");
  
  /* Note VIDIOC_S_FMT may change width and height. */
  
  /* Buggy driver paranoia. */
  min = fmt.fmt.pix.width * 2;
  if(fmt.fmt.pix.bytesperline < min)
    fmt.fmt.pix.bytesperline = min;
  min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
  if(fmt.fmt.pix.sizeimage < min)
    fmt.fmt.pix.sizeimage = min;
  
  init_mmap(img_in_mmap_buffer);

}

static void close_device(void){
  if(close(fd) == -1)
    errno_exit("close");
  
  fd = -1;
}

static void open_device(void){
  struct stat st; 
  
  if(stat(dev_name, &st) == -1) {
    fprintf(stderr, "Cannot identify '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  if(!S_ISCHR(st.st_mode)) {
    fprintf(stderr, "%s is no device\n", dev_name);
    exit(EXIT_FAILURE);
  }
  
  fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
  
  if(fd == -1) {
    fprintf(stderr, "Cannot open '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
}

static void usage(FILE *fp, int argc, char **argv){
  fprintf(fp,
	  "Usage: %s [options]\n\n"
	  "Options:\n"
	  "-c | --capture-device <name> Video device name [/dev/video0]\n"
	  "-h | --help  Print this message\n"
	  "-n | --buffer <n>  Number of images in ring buffer [25]\n"
	  "                   Enables motion detection\n"
	  "-s | --stack-length <s>  Length of stacked period in seconds [60]\n"
	  "-m | --min  Enable minimum stacking [default: off]\n"
	  "-M | --max  Enable maximum stacking [default: off]\n"
	  "-a | --ave  Enable average stacking [default: off]\n"
	  "-d | --dif  Enable difference stacking [default: off]\n"
	  "-t | --trig  Enable motion detection with default ring buffer [default: off]\n"
	  "-e | --latest  Save \"latest.png\", an independent peak-hold stack\n"
	  "-T | --minimum-trigger-level  Minimum change allowed to trigger motion\n"
	  "                              detection [10]\n"
	  "-H | --md-histogram-adjust  Motion detection adjustment [default: 10]\n"
	  "-U | --update-interval  Interval between motion detection background field\n"
	  "                        calculation in seconds [default: 60]\n"
	  "-k | --mask  Mask filename used to exclude areas from motion detect\n"
	  "-G | --trigger-test  Test how sensitive triggering is. No images are saved\n"
	  "-l | --capture-length <time>  Length of imaging period in\n"
	  "                              seconds [unlimited]\n"
	  "-L | --latest-length <time>  Length of imaging period for \"latest.png\" in\n"
	  "                             seconds [60]\n"
	  "-p | --prefix <prefix>  Image filename prefix [default: none]\n"
	  "-o | --out-dir <dir>  Image output directory [default: current directory]\n"
	  "                      The images are saved to <dir>/yyyy/mm/dd/\n"
	  "-O | --latest-fname <fname>  Filename with full path for saving\n"
	  "                             \"latest.png\" [latest.png]\n"
	  "-N | --no-log  Turn of logging\n"
	  "-C | --number_of_channels  Number of color channels (1 or 3) [default: 1]\n"
	  "\n",
	  argv[0]);
}

int main(int argc, char **argv){
  dev_name = "/dev/video0";

  int i;

  for(;;) {
    int index;
    int c;
    
    c = getopt_long(argc, argv,
		    short_options, long_options,
		    &index);
    
    if(c == -1)
      break;
    
    switch(c) {
    case 0: /* getopt_long() flag */
      break;
    case 'c':
      dev_name = optarg;
      break;      
    case 'h':
      usage(stdout, argc, argv);
      exit(EXIT_SUCCESS);
    case 'n':
      sscanf(optarg,"%d",&img_in_md_buffer);
      use_trig = 1;
      break;
    case 's':  /* length of stacked period in secs */
      sscanf(optarg,"%d",&stack_length);
      break;
    case 'L': /* Length of "latest.png" stack */
      sscanf(optarg,"%d",&latest_length);
      break;
    case 'm': /* Enable minimum stack */
      use_min = 1;
      break;
    case 'M': /* Enable maximum stack */
      use_max = 1;
      break;
    case 'a': /* Enable average stack */
      use_ave = 1;
      break;
    case 't': /* Enable triggering (motion detection)*/
      use_trig = 1;
      break;
    case 'e': /* Enable saving of "latest.png" */
      use_latest = 1;
      break;
    case 'T': /* Minimum trigger/motion detect threshold*/
      sscanf(optarg,"%hhu",&md_min_tl);
      break;
    case 'H': /* Trigger/motion detection size */
      sscanf(optarg,"%hhu",&md_histogram_adj);
      break;
    case 'U': /* Interval between background calculations */
      sscanf(optarg,"%d",&md_threshold_update_interval);
      break;
    case 'k': /* Mask that can be used to exclude pixels from motion detection */
      md_mask_fname = optarg;
      break;
    case 'G':
      md_test = 1;
      use_trig = 1;
      break;
    case 'P':
      sscanf(optarg,"%d",&md_percentage);
      break;
    case 'd': /* Enable difference stack */
      use_dif = 1;
      break;
    case 'l': /* Length of imaging period */
      sscanf(optarg,"%ld",&capture_length);
      break;
    case 'p': /* Image prefix */
      img_prefix = optarg;
      break;
    case 'o': /* Output directory prefix */
      out_dir = optarg;
      break;
    case 'O': /* Filename for "latest.png" */
      latest_fname = optarg;
      use_latest = 1;
      break;
    case 'N': /* Disable logging */
      save_log = 0;
      break;
    case 'C':
      sscanf(optarg,"%d",&number_of_channels);
      break;
    case 'V': /* Verbose */
      verbose = 1;
      break;
    default:
      usage(stderr, argc, argv);
      exit(EXIT_FAILURE);
    }
  }

  if(!use_min && !use_max && !use_ave && !use_trig && !use_dif && !use_latest){
    printf("\nNothing to do, terminating.\n");
    printf("Specify at least one imaging mode!\n\n");
    usage(stderr, argc, argv);
    exit(EXIT_FAILURE);
  }

  if(md_test){
    use_min = 0;
    use_max = 0;
    use_ave = 0;
    use_dif = 0;
    use_latest = 0;
  }

  if(number_of_channels != 1 && number_of_channels != 3){
    number_of_channels = 1;
  }
  if(number_of_channels == 3){
    init_yuyv_lookups();
  }

  if(use_latest && latest_fname == NULL){
    latest_fname = (char*)malloc(10*sizeof(char));
    strcpy(latest_fname,latest_fname_default);
  }

  ave_stack1.img_buffer = NULL;
  ave_stack1.min_val = ULONG_MAX;
  ave_stack1.max_val = 0;

  ave_stack2.img_buffer = NULL;
  ave_stack2.min_val = ULONG_MAX;
  ave_stack2.max_val = 0;

  ave_stack = &ave_stack1;
  ave_stack_save = NULL;

  peak_hold1.img_buffer = NULL;
  peak_hold2.img_buffer = NULL;
  peak_hold = &peak_hold1;
  peak_hold_save = NULL;

  min_hold1.img_buffer = NULL;
  min_hold2.img_buffer = NULL;
  min_hold = &min_hold1;
  min_hold_save = NULL;

  latest_img1.img_buffer = NULL;
  latest_img2.img_buffer = NULL;
  latest_img = &latest_img1;
  latest_img_save = NULL;

  if(use_trig){
    md_frames = calloc(img_in_md_buffer,sizeof(*md_frames));
    for(i=0; i<img_in_md_buffer; i++){
      md_frames[i].img_buffer = NULL;
      md_frames[i].luminance = NULL;
      md_frames[i].channels = number_of_channels;
      md_frames[i].width = -1;
      md_frames[i].height = -1;
      md_frames[i].npix = -1;
    }
    
    if(md_mask_fname != NULL){
      md_mask = read_mask(md_mask_fname, &md_mask_width, &md_mask_height);
    }

  }

  open_device();

  init_device(img_in_mmap_buffer);

  start_capturing();

  mainloop();

  /* Wait until all threads have finished */
  while(md_process_on || md_reference_calc_on || md_saving_images || saving_stacks || saving_latest){
  }

  if(capture_length != 0){
    stop_capturing();

    uninit_device();

    close_device();
  }

  if(use_trig){
    for(i=0; i < img_in_md_buffer; i++){
      if(md_frames[i].img_buffer != NULL){
	free(md_frames[i].img_buffer);
	md_frames[i].img_buffer = NULL;
      }
    }
  }

  if(md_frames != NULL){ free(md_frames); }
  if(peak_hold1.img_buffer != NULL){ free(peak_hold1.img_buffer); }
  if(peak_hold2.img_buffer != NULL){ free(peak_hold2.img_buffer); }
  if(min_hold1.img_buffer != NULL){ free(min_hold1.img_buffer); }
  if(min_hold2.img_buffer != NULL){ free(min_hold2.img_buffer); }
  if(ave_stack1.img_buffer != NULL){ free(ave_stack1.img_buffer); }
  if(ave_stack2.img_buffer != NULL){ free(ave_stack2.img_buffer); }
  if(latest_img1.img_buffer != NULL){ free(latest_img1.img_buffer); }
  if(latest_img2.img_buffer != NULL){ free(latest_img2.img_buffer); }

  if(md_median1 != NULL){ free(md_median1); }
  if(md_median2 != NULL){ free(md_median2); }
  if(md_luminance1 != NULL){
    for(i=0; i<img_in_md_buffer; i++){ free(md_luminance1[i]); }
    free(md_luminance1);
  }
  if(md_luminance2 != NULL){
    for(i=0; i<img_in_md_buffer; i++){ free(md_luminance2[i]); }
    free(md_luminance2);
  }
  if(md_ts_histogram1 != NULL){ free(md_ts_histogram1); }
  if(md_ts_histogram1 != NULL){ free(md_ts_histogram1); }

  if(current_frame.img_buffer != NULL){ free(current_frame.img_buffer); }
  if(current_frame.luminance != NULL){ free(current_frame.luminance); }

  if(md_log_fid != NULL){ fclose(md_log_fid); }
  if(stack_log_fid != NULL){ fclose(stack_log_fid); }

  if(md_mask != NULL){ free(md_mask); }

  exit(EXIT_SUCCESS);

  return 0;
}
